@@ -47,7 +47,12 @@ class ARICLI:
         parser.add_argument("--role-name", nargs="?", help="if provided, use it as a role name")
         parser.add_argument("--source", help="source server name in ansible config file (if empty, use public ansible galaxy)")
         parser.add_argument("--without-ram", action="store_true", help="if true, RAM data is not used and not even updated")
+        parser.add_argument("--read-only-ram", action="store_true", help="if true, RAM data is used but not updated")
+        parser.add_argument("--read-ram-for-dependency", action="store_true", help="if true, RAM data is used only for dependency")
         parser.add_argument("--update-ram", action="store_true", help="if true, RAM data is not used for scan but updated with the scan result")
+        parser.add_argument("--include-tests", action="store_true", help='if true, load test contents in "tests/integration/targets"')
+        parser.add_argument("--silent", action="store_true", help='if true, do not print anything"')
+        parser.add_argument("--objects", action="store_true", help="if true, output objects.json to the output directory")
         parser.add_argument("--show-all", action="store_true", help="if true, show findings even if missing dependencies are found")
         parser.add_argument("--json", help="if specified, show findings in json format")
         parser.add_argument("--yaml", help="if specified, show findings in yaml format")
@@ -97,7 +102,7 @@ class ARICLI:
         elif args.rules_dir:
             rules_dir = args.rules_dir + ":" + config.rules_dir
 
-        silent = False
+        silent = args.silent
         pretty = False
         output_format = ""
         if args.json or args.yaml:
@@ -110,12 +115,20 @@ class ARICLI:
 
         read_ram = True
         write_ram = True
+        read_ram_for_dependency = False
         if args.without_ram:
             read_ram = False
             write_ram = False
+        elif args.read_only_ram:
+            read_ram = True
+            write_ram = False
         elif args.update_ram:
             read_ram = False
             write_ram = True
+        elif args.read_ram_for_dependency:
+            read_ram_for_dependency = True
+            read_ram = False
+            write_ram = False
 
         c = ARIScanner(
             root_dir=config.data_dir,
@@ -123,6 +136,7 @@ class ARICLI:
             do_save=args.save,
             read_ram=read_ram,
             write_ram=write_ram,
+            read_ram_for_dependency=read_ram_for_dependency,
             show_all=args.show_all,
             silent=silent,
             pretty=pretty,
@@ -144,5 +158,7 @@ class ARICLI:
             source_repository=args.source,
             playbook_only=args.playbook_only,
             taskfile_only=args.taskfile_only,
+            include_test_contents=args.include_tests,
+            objects=args.objects,
             out_dir=args.out_dir,
         )
