@@ -24,17 +24,31 @@ from .models import (
     LoadType,
     ObjectList,
     ExecutableType,
+    Collection,
     Module,
     ModuleMetadata,
+    Role,
+    RoleMetadata,
+    TaskFile,
+    TaskFileMetadata,
 )
 from .findings import Findings
-from .utils import escape_url, version_to_num, diff_files_data
+from .utils import (
+    escape_url,
+    version_to_num,
+    diff_files_data,
+    is_test_object,
+    lock_file,
+    unlock_file,
+)
 from .safe_glob import safe_glob
-from .keyutil import get_obj_info_by_key
+from .keyutil import get_obj_info_by_key, make_imported_taskfile_key
 from .model_loader import load_builtin_modules
 
 
 module_index_name = "module_index.json"
+role_index_name = "role_index.json"
+taskfile_index_name = "taskfile_index.json"
 
 
 @dataclass
@@ -42,22 +56,22 @@ class RAMClient(object):
     root_dir: str = ""
 
     findings_json_list_cache: list = field(default_factory=list)
-    modules_json_list_cache: list = field(default_factory=list)
-    roles_json_list_cache: list = field(default_factory=list)
-    taskfiles_json_list_cache: list = field(default_factory=list)
-    tasks_json_list_cache: list = field(default_factory=list)
 
-    modules_cache: dict = field(default_factory=dict)
-    roles_cache: dict = field(default_factory=dict)
-    taskfiles_cache: dict = field(default_factory=dict)
-    tasks_cache: dict = field(default_factory=dict)
+    findings_cache: dict = field(default_factory=dict)
+    findings_search_cache: dict = field(default_factory=dict)
 
     module_search_cache: dict = field(default_factory=dict)
+    role_search_cache: dict = field(default_factory=dict)
+    taskfile_search_cache: dict = field(default_factory=dict)
     task_search_cache: dict = field(default_factory=dict)
 
     builtin_modules_cache: dict = field(default_factory=dict)
 
     module_index: dict = field(default_factory=dict)
+    role_index: dict = field(default_factory=dict)
+    taskfile_index: dict = field(default_factory=dict)
+
+    max_cache_size: int = 200
 
     def __post_init__(self):
         module_index_path = os.path.join(self.root_dir, "indices", module_index_name)
@@ -65,6 +79,35 @@ class RAMClient(object):
             with open(module_index_path, "r") as file:
                 self.module_index = json.load(file)
 
+        role_index_path = os.path.join(self.root_dir, "indices", role_index_name)
+        if os.path.exists(role_index_path):
+            with open(role_index_path, "r") as file:
+                self.role_index = json.load(file)
+
+        taskfile_index_path = os.path.join(self.root_dir, "indices", taskfile_index_name)
+        if os.path.exists(taskfile_index_path):
+            with open(taskfile_index_path, "r") as file:
+                self.taskfile_index = json.load(file)
+
+    def clear_old_cache(self):
+        size = self.max_cache_size
+        self._remove_old_item(self.findings_cache, size)
+        self._remove_old_item(self.findings_search_cache, size)
+        self._remove_old_item(self.module_search_cache, size)
+        self._remove_old_item(self.role_search_cache, size)
+        self._remove_old_item(self.taskfile_search_cache, size)
+        self._remove_old_item(self.task_search_cache, size)
+        return
+
+    def _remove_old_item(self, data: dict, size: int):
+        if len(data) <= size:
+            return
+        num = len(data) - size
+        for _ in range(num):
+            oldest_key = next(iter(data))
+            data.pop(oldest_key)
+        return
+
     def register(self, findings: Findings):
         metadata = findings.metadata
 
@@ -76,17 +119,128 @@ class RAMClient(object):
         out_dir = self.make_findings_dir_path(type, name, version, hash)
         self.save_findings(findings, out_dir)
 
-    def module_index_register(self, findings: Findings):
+        self.clear_old_cache()
+
+    def register_indices_to_ram(self, findings: Findings, include_test_contents: bool = False):
+        self.register_module_index_to_ram(findings=findings, include_test_contents=include_test_contents)
+        self.register_role_index_to_ram(findings=findings, include_test_contents=include_test_contents)
+        self.register_taskfile_index_to_ram(findings=findings, include_test_contents=include_test_contents)
+
+    def register_module_index_to_ram(self, findings: Findings, include_test_contents: bool = False):
+        new_data_found = False
         modules = self.load_module_index()
         for module in findings.root_definitions.get("definitions", {}).get("modules", []):
             if not isinstance(module, Module):
                 continue
+            if include_test_contents and is_test_object(module.defined_in):
+                continue
             m_meta = ModuleMetadata.from_module(module, findings.metadata)
             current = modules.get(module.name, [])
-            if m_meta not in current:
+            exists = False
+            for m_dict in current:
+                m = None
+                if isinstance(m_dict, dict):
+                    m = ModuleMetadata.from_dict(m_dict)
+                elif isinstance(m_dict, ModuleMetadata):
+                    m = m_dict
+                if not m:
+                    continue
+                if m == m_meta:
+                    exists = True
+                    break
+            if not exists:
                 current.append(m_meta)
+                new_data_found = True
             modules.update({module.name: current})
-        self.save_module_index(modules)
+        for collection in findings.root_definitions.get("definitions", {}).get("collections", []):
+            if not isinstance(collection, Collection):
+                continue
+            if collection.meta_runtime and isinstance(collection.meta_runtime, dict):
+                for short_name, routing in collection.meta_runtime.get("plugin_routing", {}).get("modules", {}).items():
+                    redirect_to = routing.get("redirect", "")
+                    if not redirect_to:
+                        continue
+                    m_meta = ModuleMetadata.from_routing(redirect_to, findings.metadata)
+                    current = modules.get(short_name, [])
+                    exists = False
+                    for m_dict in current:
+                        m = None
+                        if isinstance(m_dict, dict):
+                            m = ModuleMetadata.from_dict(m_dict)
+                        elif isinstance(m_dict, ModuleMetadata):
+                            m = m_dict
+                        if not m:
+                            continue
+                        if m == m_meta:
+                            exists = True
+                            break
+                    if not exists:
+                        current.append(m_meta)
+                        new_data_found = True
+                    modules.update({short_name: current})
+        if new_data_found:
+            self.save_module_index(modules)
+        return
+
+    def register_role_index_to_ram(self, findings: Findings, include_test_contents: bool = False):
+        new_data_found = False
+        roles = self.load_role_index()
+        for role in findings.root_definitions.get("definitions", {}).get("roles", []):
+            if not isinstance(role, Role):
+                continue
+            if include_test_contents and is_test_object(role.defined_in):
+                continue
+            r_meta = RoleMetadata.from_role(role, findings.metadata)
+            current = roles.get(r_meta.fqcn, [])
+            exists = False
+            for r_dict in current:
+                r = None
+                if isinstance(r_dict, dict):
+                    r = RoleMetadata.from_dict(r_dict)
+                elif isinstance(r_dict, RoleMetadata):
+                    r = r_dict
+                if not r:
+                    continue
+                if r == r_meta:
+                    exists = True
+                    break
+            if not exists:
+                current.append(r_meta)
+                new_data_found = True
+            roles.update({role.fqcn: current})
+        if new_data_found:
+            self.save_role_index(roles)
+        return
+
+    def register_taskfile_index_to_ram(self, findings: Findings, include_test_contents: bool = False):
+        new_data_found = False
+        taskfiles = self.load_taskfile_index()
+        for taskfile in findings.root_definitions.get("definitions", {}).get("taskfiles", []):
+            if not isinstance(taskfile, TaskFile):
+                continue
+            if include_test_contents and is_test_object(taskfile.defined_in):
+                continue
+            tf_meta = TaskFileMetadata.from_taskfile(taskfile, findings.metadata)
+            current = taskfiles.get(tf_meta.key, [])
+            exists = False
+            for tf_dict in current:
+                tf = None
+                if isinstance(tf_dict, dict):
+                    tf = TaskFileMetadata.from_dict(tf_dict)
+                elif isinstance(tf_dict, TaskFileMetadata):
+                    tf = tf_dict
+                if not tf:
+                    continue
+                if tf == tf_meta:
+                    exists = True
+                    break
+            if not exists:
+                current.append(tf_meta)
+                new_data_found = True
+            taskfiles.update({taskfile.key: current})
+        if new_data_found:
+            self.save_taskfile_index(taskfiles)
+        return
 
     def make_findings_dir_path(self, type, name, version, hash):
         type_root = type + "s"
@@ -224,14 +378,15 @@ class RAMClient(object):
         search_end = False
         for findings_json in modules_json_list:
             modules = ObjectList()
-            if findings_json in self.modules_cache:
-                modules = self.modules_cache[findings_json]
+            if findings_json in self.findings_cache:
+                modules = self.findings_cache[findings_json].get("modules", [])
             else:
                 f = Findings.load(fpath=findings_json)
                 if not isinstance(f, Findings):
                     continue
-                modules = f.root_definitions.get("definitions", {}).get("modules", [])
-                self.modules_cache[findings_json] = modules
+                definitions = f.root_definitions.get("definitions", {})
+                modules = definitions.get("modules", [])
+                self.findings_cache[findings_json] = definitions
             for m in modules:
                 matched = False
                 if exact_match:
@@ -266,55 +421,47 @@ class RAMClient(object):
         self.module_search_cache[args_str] = matched_modules
         return matched_modules
 
-    def search_role(self, name, exact_match=False, max_match=-1, collection_name="", collection_version="", used_in=""):
+    def search_role(self, name, exact_match=False, max_match=-1, used_in=""):
         if max_match == 0:
             return []
-        findings_json_list = []
-        if self.findings_json_list_cache:
-            findings_json_list = self.findings_json_list_cache
-        else:
-            search_patterns = os.path.join(self.root_dir, "collections", "findings", "*", "*", "*", "findings.json")
-            findings_json_list_coll = safe_glob(search_patterns)
-            findings_json_list_coll = sort_by_version(findings_json_list_coll)
-            search_patterns = os.path.join(self.root_dir, "roles", "findings", "*", "*", "*", "findings.json")
-            findings_json_list_role = safe_glob(search_patterns)
-            findings_json_list_role = sort_by_version(findings_json_list_role)
-            findings_json_list = findings_json_list_coll + findings_json_list_role
-            self.findings_json_list_cache = findings_json_list
+        args_str = json.dumps([name, exact_match, max_match])
+        if args_str in self.role_search_cache:
+            return self.role_search_cache[args_str]
+
+        from_indices = False
+        found_index = None
+        if name in self.role_index and self.role_index[name]:
+            from_indices = True
+            found_index = self.role_index[name][0]
 
         roles_json_list = []
-        if self.roles_json_list_cache:
-            roles_json_list = self.roles_json_list_cache
+        if from_indices:
+            _type = found_index.get("type", "")
+            _name = found_index.get("name", "")
+            _version = found_index.get("version", "")
+            _hash = found_index.get("hash", "")
+            findings_path = os.path.join(self.root_dir, _type + "s", "findings", _name, _version, _hash, "findings.json")
+            if os.path.exists(findings_path):
+                roles_json_list.append(findings_path)
         else:
-            for findings_json in findings_json_list:
-                f = Findings.load(fpath=findings_json)
-                if not isinstance(f, Findings):
-                    continue
-                # avoid using unresolved RAM data
-                if f.extra_requirements:
-                    continue
-                roles = f.root_definitions.get("definitions", {}).get("roles", [])
-                self.roles_cache[findings_json] = roles
-                roles_json_list.append(findings_json)
-            self.roles_json_list_cache = roles_json_list
-
-        if collection_name != "":
-            roles_json_list = [fpath for fpath in roles_json_list if f"/{collection_name}/" in fpath]
-        if collection_version != "":
-            roles_json_list = [fpath for fpath in roles_json_list if f"/{collection_version}/" in fpath]
+            # Do not search a role from all findings
+            # when it is not found in the role index.
+            # Instead, just return nothing in the case.
+            pass
 
         matched_roles = []
         search_end = False
         for findings_json in roles_json_list:
             roles = ObjectList()
-            if findings_json in self.roles_cache:
-                roles = self.roles_cache[findings_json]
+            if findings_json in self.findings_cache:
+                roles = self.findings_cache[findings_json].get("roles", [])
             else:
                 f = Findings.load(fpath=findings_json)
                 if not isinstance(f, Findings):
                     continue
-                roles = f.root_definitions.get("definitions", {}).get("roles", [])
-                self.roles_cache[findings_json] = roles
+                definitions = f.root_definitions.get("definitions", {})
+                roles = definitions.get("roles", [])
+                self.findings_cache[findings_json] = definitions
             for r in roles:
                 matched = False
                 if exact_match:
@@ -327,7 +474,7 @@ class RAMClient(object):
                     parts = findings_json.split("/")
                     offspring_objects = []
                     for taskfile_key in r.taskfiles:
-                        _tmp_offspring_objects = self.search_taskfile(taskfile_key, is_key=True, collection_name=r.collection)
+                        _tmp_offspring_objects = self.search_taskfile(taskfile_key, is_key=True)
                         if len(_tmp_offspring_objects) > 0:
                             tf = _tmp_offspring_objects[0]
                             if tf:
@@ -363,82 +510,93 @@ class RAMClient(object):
                         break
             if search_end:
                 break
+        self.role_search_cache[args_str] = matched_roles
         return matched_roles
 
-    def search_taskfile(self, name, include_task_path="", max_match=-1, is_key=False, collection_name="", collection_version="", used_in=""):
-        if max_match == 0:
-            return []
-        findings_json_list = []
-        if self.findings_json_list_cache:
-            findings_json_list = self.findings_json_list_cache
-        else:
-            search_patterns = os.path.join(self.root_dir, "collections", "findings", "*", "*", "*", "findings.json")
-            findings_json_list_coll = safe_glob(search_patterns)
-            findings_json_list_coll = sort_by_version(findings_json_list_coll)
-            search_patterns = os.path.join(self.root_dir, "roles", "findings", "*", "*", "*", "findings.json")
-            findings_json_list_role = safe_glob(search_patterns)
-            findings_json_list_role = sort_by_version(findings_json_list_role)
-            findings_json_list = findings_json_list_coll + findings_json_list_role
-            self.findings_json_list_cache = findings_json_list
-
-        taskfiles_json_list = []
-        if self.taskfiles_json_list_cache:
-            taskfiles_json_list = self.taskfiles_json_list_cache
-        else:
-            for findings_json in findings_json_list:
-                f = Findings.load(fpath=findings_json)
-                if not isinstance(f, Findings):
-                    continue
-                # avoid using unresolved RAM data
-                if f.extra_requirements:
-                    continue
-                taskfiles = f.root_definitions.get("definitions", {}).get("taskfiles", [])
-                self.taskfiles_cache[findings_json] = taskfiles
-                taskfiles_json_list.append(findings_json)
-            self.taskfiles_json_list_cache = taskfiles_json_list
-
-        if collection_name != "":
-            taskfiles_json_list = [fpath for fpath in taskfiles_json_list if f"/{collection_name}/" in fpath]
-        if collection_version != "":
-            taskfiles_json_list = [fpath for fpath in taskfiles_json_list if f"/{collection_version}/" in fpath]
-
-        search_path_list = []
+    def make_taskfile_key_candidates(self, name, from_path, from_key):
+        key_candidates = []
         taskfile_ref = name
-        if include_task_path != "":
-            base_path = os.path.dirname(include_task_path)
+        if from_path:
+            base_path = os.path.dirname(from_path)
             taskfile_path = os.path.normpath(os.path.join(base_path, taskfile_ref))
-            search_path_list.append(taskfile_path)
+            candidate_key_1 = make_imported_taskfile_key(from_key, taskfile_path)
+            key_candidates.append(candidate_key_1)
             if "roles/" in taskfile_ref and "roles/" in base_path:
                 root_path = base_path.split("roles/")[0]
                 taskfile_path = os.path.normpath(os.path.join(root_path, taskfile_ref))
-                search_path_list.append(taskfile_path)
+                candidate_key_2 = make_imported_taskfile_key(from_key, taskfile_path)
+                key_candidates.append(candidate_key_2)
+
+        return key_candidates
+
+    def search_taskfile(self, name, from_path="", from_key="", max_match=-1, is_key=False, used_in=""):
+        if max_match == 0:
+            return []
+
+        # it name is not an object key, we need `from_path` to create a key to be searched
+        if not is_key and not from_path:
+            return []
+
+        args_str = json.dumps([name, from_path, from_key, max_match, is_key])
+        if args_str in self.taskfile_search_cache:
+            return self.taskfile_search_cache[args_str]
+
+        from_indices = False
+        found_index = None
+        found_key = ""
+        taskfile_key_candidates = []
+        if is_key:
+            taskfile_key_candidates = [name]
+        else:
+            taskfile_key_candidates = self.make_taskfile_key_candidates(name, from_path, from_key)
+        for taskfile_key in taskfile_key_candidates:
+            if taskfile_key in self.taskfile_index and self.taskfile_index[taskfile_key]:
+                from_indices = True
+                found_index = self.taskfile_index[taskfile_key][0]
+                found_key = taskfile_key
+                break
+
+        taskfiles_json_list = []
+        content_info = None
+        if from_indices:
+            _type = found_index.get("type", "")
+            _name = found_index.get("name", "")
+            _version = found_index.get("version", "")
+            _hash = found_index.get("hash", "")
+            content_info = found_index
+            findings_path = os.path.join(self.root_dir, _type + "s", "findings", _name, _version, _hash, "findings.json")
+            if os.path.exists(findings_path):
+                taskfiles_json_list.append(findings_path)
+        else:
+            # Do not search a role from all findings
+            # when it is not found in the role index.
+            # Instead, just return nothing in the case.
+            pass
 
         matched_taskfiles = []
         search_end = False
         for findings_json in taskfiles_json_list:
             taskfiles = ObjectList()
-            if findings_json in self.taskfiles_cache:
-                taskfiles = self.taskfiles_cache[findings_json]
+            if findings_json in self.findings_cache:
+                taskfiles = self.findings_cache[findings_json].get("taskfiles", [])
             else:
                 f = Findings.load(fpath=findings_json)
                 if not isinstance(f, Findings):
                     continue
-                taskfiles = f.root_definitions.get("definitions", {}).get("taskfiles", [])
-                self.taskfiles_cache[findings_json] = taskfiles
+                definitions = f.root_definitions.get("definitions", {})
+                taskfiles = definitions.get("taskfiles", [])
+                self.findings_cache[findings_json] = definitions
             for tf in taskfiles:
                 matched = False
-                if is_key:
-                    if tf.key == name:
-                        matched = True
-                else:
-                    if tf.defined_in in search_path_list:
-                        matched = True
-                    # TODO: support taskfile reference with variables
+                if tf.key == found_key:
+                    matched = True
+
+                # TODO: support taskfile reference with variables
                 if matched:
                     parts = findings_json.split("/")
                     offspring_objects = []
                     for task_key in tf.tasks:
-                        _tmp_offspring_objects = self.search_task(task_key, is_key=True, collection_name=tf.collection, used_in=used_in)
+                        _tmp_offspring_objects = self.search_task(task_key, is_key=True, content_info=content_info, used_in=used_in)
                         if len(_tmp_offspring_objects) > 0:
                             t = _tmp_offspring_objects[0]
                             if t:
@@ -477,58 +635,42 @@ class RAMClient(object):
                 break
         return matched_taskfiles
 
-    def search_task(self, name, exact_match=False, max_match=-1, is_key=False, collection_name="", collection_version="", used_in=""):
+    def search_task(self, name, exact_match=False, max_match=-1, is_key=False, content_info=None, used_in=""):
         if max_match == 0:
             return []
-        args_str = json.dumps([name, exact_match, max_match, is_key, collection_name, collection_version])
+        # search task in RAM must be done for a specific content (collection/role)
+        # so give up search here when no content_info is provided
+        if not content_info or not isinstance(content_info, dict):
+            return []
+
+        args_str = json.dumps([name, exact_match, max_match, is_key, content_info])
         if args_str in self.task_search_cache:
             return self.task_search_cache[args_str]
-        findings_json_list = []
-        if self.findings_json_list_cache:
-            findings_json_list = self.findings_json_list_cache
-        else:
-            search_patterns = os.path.join(self.root_dir, "collections", "findings", "*", "*", "*", "findings.json")
-            findings_json_list_coll = safe_glob(search_patterns)
-            findings_json_list_coll = sort_by_version(findings_json_list_coll)
-            search_patterns = os.path.join(self.root_dir, "roles", "findings", "*", "*", "*", "findings.json")
-            findings_json_list_role = safe_glob(search_patterns)
-            findings_json_list_role = sort_by_version(findings_json_list_role)
-            findings_json_list = findings_json_list_coll + findings_json_list_role
-            self.findings_json_list_cache = findings_json_list
 
         tasks_json_list = []
-        if self.tasks_json_list_cache:
-            tasks_json_list = self.tasks_json_list_cache
-        else:
-            for findings_json in findings_json_list:
-                f = Findings.load(fpath=findings_json)
-                if not isinstance(f, Findings):
-                    continue
-                # avoid using unresolved RAM data
-                if f.extra_requirements:
-                    continue
-                tasks = f.root_definitions.get("definitions", {}).get("tasks", [])
-                self.tasks_cache[findings_json] = tasks
-                tasks_json_list.append(findings_json)
-            self.tasks_json_list_cache = tasks_json_list
-
-        if collection_name != "":
-            tasks_json_list = [fpath for fpath in tasks_json_list if f"/{collection_name}/" in fpath]
-        if collection_version != "":
-            tasks_json_list = [fpath for fpath in tasks_json_list if f"/{collection_version}/" in fpath]
+        _type = content_info.get("type", "")
+        if _type:
+            _type = _type + "s"
+        _name = content_info.get("name", "")
+        _version = content_info.get("version", "")
+        _hash = content_info.get("hash", "")
+        findings_path = os.path.join(self.root_dir, _type, "findings", _name, _version, _hash, "findings.json")
+        if os.path.exists(findings_path):
+            tasks_json_list.append(findings_path)
 
         matched_tasks = []
         search_end = False
         for findings_json in tasks_json_list:
             tasks = ObjectList()
-            if findings_json in self.tasks_cache:
-                tasks = self.tasks_cache[findings_json]
+            if findings_json in self.findings_cache:
+                tasks = self.findings_cache[findings_json].get("tasks", [])
             else:
                 f = Findings.load(fpath=findings_json)
                 if not isinstance(f, Findings):
                     continue
-                tasks = f.root_definitions.get("definitions", {}).get("tasks", [])
-                self.tasks_cache[findings_json] = tasks
+                definitions = f.root_definitions.get("definitions", {})
+                tasks = definitions.get("tasks", [])
+                self.findings_cache[findings_json] = definitions
             for t in tasks:
                 matched = False
                 if is_key:
@@ -549,9 +691,7 @@ class RAMClient(object):
                     elif t.executable_type == ExecutableType.ROLE_TYPE:
                         _tmp_offspring_objects = self.search_role(t.executable, used_in=t.defined_in)
                     elif t.executable_type == ExecutableType.TASKFILE_TYPE:
-                        _tmp_offspring_objects = self.search_taskfile(
-                            t.executable, include_task_path=t.defined_in, collection_name=t.collection, used_in=t.defined_in
-                        )
+                        _tmp_offspring_objects = self.search_taskfile(t.executable, from_path=t.defined_in, from_key=t.key, used_in=t.defined_in)
                     if len(_tmp_offspring_objects) > 0:
                         child = _tmp_offspring_objects[0]
                         if child:
@@ -622,19 +762,20 @@ class RAMClient(object):
                 }
         return matched_obj
 
+    def init_findings_json_list_cache(self):
+        search_patterns = os.path.join(self.root_dir, "collections", "findings", "*", "*", "*", "findings.json")
+        findings_json_list_coll = safe_glob(search_patterns)
+        findings_json_list_coll = sort_by_version(findings_json_list_coll)
+        search_patterns = os.path.join(self.root_dir, "roles", "findings", "*", "*", "*", "findings.json")
+        findings_json_list_role = safe_glob(search_patterns)
+        findings_json_list_role = sort_by_version(findings_json_list_role)
+        findings_json_list = findings_json_list_coll + findings_json_list_role
+        self.findings_json_list_cache = findings_json_list
+
     def list_all_ram_metadata(self):
-        findings_json_list = []
-        if self.findings_json_list_cache:
-            findings_json_list = self.findings_json_list_cache
-        else:
-            search_patterns = os.path.join(self.root_dir, "collections", "findings", "*", "*", "*", "findings.json")
-            findings_json_list_coll = safe_glob(search_patterns)
-            findings_json_list_coll = sort_by_version(findings_json_list_coll)
-            search_patterns = os.path.join(self.root_dir, "roles", "findings", "*", "*", "*", "findings.json")
-            findings_json_list_role = safe_glob(search_patterns)
-            findings_json_list_role = sort_by_version(findings_json_list_role)
-            findings_json_list = findings_json_list_coll + findings_json_list_role
-            self.findings_json_list_cache = findings_json_list
+        if not self.findings_json_list_cache:
+            self.init_findings_json_list_cache()
+        findings_json_list = self.findings_json_list_cache
 
         metadata_list = []
         for findings_path in findings_json_list:
@@ -651,25 +792,33 @@ class RAMClient(object):
         return metadata_list
 
     def search_findings(self, target_name, target_version, target_type=None):
+        if not self.findings_json_list_cache:
+            self.init_findings_json_list_cache()
+        args_str = json.dumps([target_name, target_version, target_type])
+        if args_str in self.findings_search_cache:
+            return self.findings_search_cache[args_str]
+
         if not target_name:
             raise ValueError("target name must be specified for searching RAM data")
         if not target_version:
             target_version = "*"
-        search_patterns = []
-        if not target_type or target_type == "collection":
-            search_patterns.append(os.path.join(self.root_dir, "collections", "findings", target_name, target_version, "*", "findings.json"))
-        if not target_type or target_type == "role":
-            search_patterns.append(os.path.join(self.root_dir, "roles", "findings", target_name, target_version, "*", "findings.json"))
-        if not target_type or target_type == "project":
-            e_target_name = escape_url(target_name)
-            search_patterns.append(os.path.join(self.root_dir, "projects", "findings", e_target_name, target_version, "*", "findings.json"))
-        if not target_type or target_type == "playbook":
-            e_target_name = escape_url(target_name)
-            search_patterns.append(os.path.join(self.root_dir, "playbooks", "findings", e_target_name, target_version, "*", "findings.json"))
-        if not target_type or target_type == "taskfile":
-            e_target_name = escape_url(target_name)
-            search_patterns.append(os.path.join(self.root_dir, "taskfiles", "findings", e_target_name, target_version, "*", "findings.json"))
-        found_path_list = safe_glob(search_patterns)
+        findings_json_list = self.findings_json_list_cache
+        found_path_list = []
+        for findings_path in findings_json_list:
+            parts = findings_path.split("/")
+            _type = parts[-5][:-1]  # collection or role
+            _name = parts[-4]
+            _version = parts[-3]
+            if _name != target_name:
+                continue
+            if target_version and target_version != "*":
+                if _version != target_name:
+                    continue
+            if target_type and target_type != "*":
+                if _type != target_type:
+                    continue
+            found_path_list.append(findings_path)
+
         latest_findings_path = ""
         if len(found_path_list) == 1:
             latest_findings_path = found_path_list[0]
@@ -684,6 +833,8 @@ class RAMClient(object):
         findings = None
         if os.path.exists(latest_findings_path):
             findings = self.load_findings(latest_findings_path)
+
+        self.findings_search_cache[args_str] = findings
         return findings
 
     def load_findings(self, path: str):
@@ -705,21 +856,41 @@ class RAMClient(object):
 
         findings.dump(fpath=os.path.join(out_dir, "findings.json"))
 
-    def save_module_index(self, modules):
+    def save_index(self, index_objects, filename):
         out_dir = os.path.join(self.root_dir, "indices")
         if not os.path.exists(out_dir):
             os.makedirs(out_dir, exist_ok=True)
-        modules_str = jsonpickle.encode(modules, make_refs=False, unpicklable=False)
-        with open(os.path.join(out_dir, "module_index.json"), "w") as file:
-            file.write(modules_str)
+        index_objects_str = jsonpickle.encode(index_objects, make_refs=False, unpicklable=False)
+        with open(os.path.join(out_dir, filename), "w") as file:
+            lock_file(file)
+            file.write(index_objects_str)
+            unlock_file(file)
 
-    def load_module_index(self):
-        path = os.path.join(self.root_dir, "indices", "module_index.json")
-        modules = {}
+    def load_index(self, filename=""):
+        path = os.path.join(self.root_dir, "indices", filename)
+        index_objects = {}
         if os.path.exists(path):
             with open(path, "r") as file:
-                modules = json.load(file)
-        return modules
+                index_objects = json.load(file)
+        return index_objects
+
+    def save_module_index(self, modules):
+        return self.save_index(modules, module_index_name)
+
+    def load_module_index(self):
+        return self.load_index(module_index_name)
+
+    def save_role_index(self, roles):
+        return self.save_index(roles, role_index_name)
+
+    def load_role_index(self):
+        return self.load_index(role_index_name)
+
+    def save_taskfile_index(self, taskfiles):
+        return self.save_index(taskfiles, taskfile_index_name)
+
+    def load_taskfile_index(self):
+        return self.load_index(taskfile_index_name)
 
     def save_error(self, error: str, out_dir: str):
         if out_dir == "":
@@ -762,7 +933,7 @@ class RAMClient(object):
         indices = os.path.join(self.root_dir, "indices")
         collection_findings = os.path.join(self.root_dir, "collections", "findings")
         role_findings = os.path.join(self.root_dir, "roles", "findings")
-        with tarfile.open(outfile, 'w:gz') as tar:
+        with tarfile.open(outfile, "w:gz") as tar:
             if os.path.exists(indices):
                 tar.add(indices, arcname="indices")
             if os.path.exists(collection_findings):
