@@ -24,6 +24,7 @@ import datetime
 from dataclasses import dataclass, field
 
 from .models import (
+    Object,
     Load,
     LoadType,
     ObjectList,
@@ -99,7 +100,10 @@ class Config:
         if not self.disable_default_rules:
             self.disable_default_rules = self._get_single_config("ARI_DISABLE_DEFAULT_RULES", "disable_default_rules", default_disable_default_rules)
         if not self.rules_dir:
-            self.rules_dir = self._get_single_config("ARI_RULES_DIR", "rules_dir", default_rules_dir)
+            if self.disable_default_rules:
+                self.rules_dir = self._get_single_config("ARI_RULES_DIR", "rules_dir", "")
+            else:
+                self.rules_dir = self._get_single_config("ARI_RULES_DIR", "rules_dir", default_rules_dir)
         # automatically add the default rules dir unless it is disabled
         if not self.rules_dir.endswith(default_rules_dir) and not self.disable_default_rules:
             self.rules_dir += ":" + default_rules_dir
@@ -111,13 +115,17 @@ class Config:
             self.rules = self._get_single_config("ARI_RULES", "rules", default_rules, "list", ",")
 
     def _get_single_config(self, env_key: str = "", yaml_key: str = "", __default: any = None, __type=None, separator=""):
-        _from_env = os.environ.get(env_key, None)
-        if _from_env and __type:
-            if __type == "list":
-                _from_env = _from_env.split(separator)
-        _from_file = self._data.get(yaml_key, None)
-        value = _from_env or _from_file or __default
-        return value
+        if env_key in os.environ:
+            _from_env = os.environ.get(env_key, None)
+            if _from_env and __type:
+                if __type == "list":
+                    _from_env = _from_env.split(separator)
+            return _from_env
+        elif yaml_key in self._data:
+            _from_file = self._data.get(yaml_key, None)
+            return _from_file
+        else:
+            return __default
 
 
 collection_manifest_json = "MANIFEST.json"
@@ -162,6 +170,8 @@ class SingleScan(object):
     root_definitions: dict = field(default_factory=dict)
     ext_definitions: dict = field(default_factory=dict)
 
+    target_object: Object = field(default_factory=Object)
+
     trees: list = field(default_factory=list)
     # for inventory object
     additional: ObjectList = field(default_factory=ObjectList)
@@ -189,6 +199,8 @@ class SingleScan(object):
     source_repository: str = ""
     out_dir: str = ""
 
+    include_test_contents: bool = False
+
     extra_requirements: list = field(default_factory=list)
     resolve_failures: dict = field(default_factory=dict)
 
@@ -199,6 +211,7 @@ class SingleScan(object):
     root_dir: str = ""
     rules_dir: str = ""
     rules: list = field(default_factory=list)
+    persist_dependency_cache: bool = False
     spec_mutations_from_previous_scan: dict = field(default_factory=dict)
     spec_mutations: dict = field(default_factory=dict)
     use_ansible_doc: bool = True
@@ -364,6 +377,7 @@ class SingleScan(object):
             do_save=self.do_save,
             silent=self.silent,
             tmp_install_dir=self.tmp_install_dir,
+            periodical_cleanup=self.persist_dependency_cache,
         )
         dep_dirs = ddp.prepare_dir(
             root_install=root_install,
@@ -400,6 +414,7 @@ class SingleScan(object):
             playbook_only=self.playbook_only,
             taskfile_yaml=self.taskfile_yaml,
             taskfile_only=self.taskfile_only,
+            include_test_contents=self.include_test_contents,
         )
         load_object(ld)
         return ld
@@ -515,6 +530,22 @@ class SingleScan(object):
                     self.root_definitions["definitions"][type_name][i] = mutated_spec
         return
 
+    def set_target_object(self):
+        type_name = self.type + "s"
+        obj_list = self.root_definitions.get("definitions", {}).get(type_name, [])
+        if len(obj_list) == 0:
+            return
+        elif len(obj_list) == 1:
+            self.target_object = obj_list[0]
+        else:
+            # only for playbook / taskfile not in `--xxxx-only` mode
+            for obj in obj_list:
+                obj_path = getattr(obj, "defined_in")
+                if self.name in obj_path:
+                    self.target_object = obj
+                    break
+        return
+
     def construct_trees(self, ram_client=None):
         trees, additional, extra_requirements, resolve_failures = tree(
             self.root_definitions, self.ext_definitions, ram_client, self.target_playbook_name, self.target_taskfile_name
@@ -558,8 +589,9 @@ class SingleScan(object):
         taskcalls_in_trees = resolve(self.trees, self.additional)
         self.taskcalls_in_trees = taskcalls_in_trees
 
-        for tree in self.trees:
-            ctx = AnsibleRunContext.from_tree(tree)
+        for i, tree in enumerate(self.trees):
+            last_item = i + 1 == len(self.trees)
+            ctx = AnsibleRunContext.from_tree(tree=tree, parent=self.target_object, last_item=last_item)
             self.contexts.append(ctx)
 
         if self.do_save:
@@ -694,7 +726,10 @@ class ARIScanner(object):
 
     ram_client: RAMClient = None
     read_ram: bool = True
-    write_ram: bool = True
+    read_ram_for_dependency: bool = True
+    write_ram: bool = False
+
+    persist_dependency_cache: bool = False
 
     skip_playbook_format_error: bool = (True,)
     skip_task_format_error: bool = (True,)
@@ -745,6 +780,8 @@ class ARIScanner(object):
         target_path: str = "",
         dependency_dir: str = "",
         download_only: bool = False,
+        load_only: bool = False,
+        skip_dependency: bool = False,
         use_src_cache: bool = False,
         source_repository: str = "",
         playbook_yaml: str = "",
@@ -752,6 +789,8 @@ class ARIScanner(object):
         taskfile_yaml: str = "",
         taskfile_only: bool = False,
         raw_yaml: str = "",
+        include_test_contents: bool = False,
+        objects: bool = False,
         out_dir: str = "",
         spec_mutations_from_previous_scan: dict = None,
     ):
@@ -788,10 +827,12 @@ class ARIScanner(object):
             playbook_only=playbook_only,
             taskfile_yaml=taskfile_yaml,
             taskfile_only=taskfile_only,
+            include_test_contents=include_test_contents,
             out_dir=out_dir,
             root_dir=self.root_dir,
             rules_dir=self.rules_dir,
             rules=self.rules,
+            persist_dependency_cache=self.persist_dependency_cache,
             spec_mutations_from_previous_scan=spec_mutations_from_previous_scan,
             use_ansible_doc=self.use_ansible_doc,
             do_save=self.do_save,
@@ -803,7 +844,7 @@ class ARIScanner(object):
 
         self.record_begin(time_records, "metadata_load")
         metdata_loaded = False
-        if self.read_ram:
+        if self.read_ram and scandata.type not in [LoadType.PLAYBOOK, LoadType.TASKFILE, LoadType.PROJECT]:
             loaded, metadata, dependencies = self.load_metadata_from_ram(scandata.type, scandata.name, scandata.version)
             logger.debug(f"metadata loaded: {loaded}")
             if loaded:
@@ -813,9 +854,9 @@ class ARIScanner(object):
                     logger.debug(f'Use metadata for "{scandata.name}" in RAM DB')
 
         if scandata.install_dependencies and not metdata_loaded:
-            print(f"start preparing {scandata.type} {scandata.name}")
+            logger.debug(f"start preparing {scandata.type} {scandata.name}")
             scandata._prepare_dependencies()
-            print(f"finished preparing {scandata.type} {scandata.name}")
+            logger.debug(f"finished preparing {scandata.type} {scandata.name}")
 
         if download_only:
             scandata.set_metadata_findings()
@@ -823,20 +864,83 @@ class ARIScanner(object):
             return None
         self.record_end(time_records, "metadata_load")
 
-        ext_list = []
-        ext_list.extend(
-            [
-                (
-                    d.get("metadata", {}).get("type", ""),
-                    d.get("metadata", {}).get("name", ""),
-                    d.get("metadata", {}).get("version", ""),
-                    d.get("metadata", {}).get("hash", ""),
-                    d.get("dir"),
-                )
-                for d in scandata.loaded_dependency_dirs
-            ]
-        )
-        ext_count = len(ext_list)
+        if not skip_dependency:
+            ext_list = []
+            ext_list.extend(
+                [
+                    (
+                        d.get("metadata", {}).get("type", ""),
+                        d.get("metadata", {}).get("name", ""),
+                        d.get("metadata", {}).get("version", ""),
+                        d.get("metadata", {}).get("hash", ""),
+                        d.get("dir"),
+                    )
+                    for d in scandata.loaded_dependency_dirs
+                ]
+            )
+            ext_count = len(ext_list)
+
+            # Start ARI Scanner main flow
+            self.record_begin(time_records, "dependency_load")
+            for i, (ext_type, ext_name, ext_ver, ext_hash, ext_path) in enumerate(ext_list):
+                if not self.silent:
+                    if i == 0:
+                        logger.info("start loading {} {}(s)".format(ext_count, ext_type))
+                    logger.info("[{}/{}] {} {}".format(i + 1, ext_count, ext_type, ext_name))
+
+                # avoid infinite loop
+                is_root = False
+                if scandata.type == ext_type and scandata.name == ext_name:
+                    is_root = True
+
+                if not is_root:
+                    key = "{}-{}".format(ext_type, ext_name)
+                    read_ram_for_dependency = self.read_ram or self.read_ram_for_dependency
+
+                    dep_loaded = False
+                    if read_ram_for_dependency:
+                        # searching findings from ARI RAM and use them if found
+                        dep_loaded, ext_defs = self.load_definitions_from_ram(ext_type, ext_name, ext_ver, ext_hash)
+                        if dep_loaded:
+                            scandata.ext_definitions[key] = ext_defs
+                            if not self.silent:
+                                logger.debug(f'Use spec data for "{ext_name}" in RAM DB')
+
+                    if not dep_loaded:
+                        # scan dependencies and save findings to ARI RAM
+                        dep_scanner = ARIScanner(
+                            root_dir=self.root_dir,
+                            rules_dir="",
+                            rules=[],
+                            ram_client=self.ram_client,
+                            read_ram=read_ram_for_dependency,
+                            read_ram_for_dependency=self.read_ram_for_dependency,
+                            write_ram=self.write_ram,
+                            use_ansible_doc=self.use_ansible_doc,
+                            do_save=self.do_save,
+                            silent=True,
+                        )
+                        # use prepared dep dirs
+                        dep_scanner.evaluate(
+                            type=ext_type,
+                            name=ext_name,
+                            version=ext_ver,
+                            hash=ext_hash,
+                            target_path=ext_path,
+                            dependency_dir=scandata.dependency_dir,
+                            skip_dependency=True,
+                            source_repository=scandata.source_repository,
+                            include_test_contents=include_test_contents,
+                            load_only=True,
+                        )
+                        dep_scandata = dep_scanner.get_last_scandata()
+                        scandata.ext_definitions[key] = dep_scandata.root_definitions
+                        dep_loaded = True
+
+            self.record_end(time_records, "dependency_load")
+
+            if not self.silent:
+                logger.debug("load_definition_ext() done")
 
         # PRM Finder
         self.record_begin(time_records, "prm_load")
@@ -846,63 +950,9 @@ class ARIScanner(object):
         # scandata.prm["modules"] = modules
         self.record_end(time_records, "prm_load")
 
-        # Start ARI Scanner main flow
-        self.record_begin(time_records, "dependency_load")
-        for i, (ext_type, ext_name, ext_ver, ext_hash, ext_path) in enumerate(ext_list):
-            if not self.silent:
-                if i == 0:
-                    logger.info("start loading {} {}(s)".format(ext_count, ext_type))
-                logger.info("[{}/{}] {} {}".format(i + 1, ext_count, ext_type, ext_name))
-
-            # avoid infinite loop
-            is_root = False
-            if scandata.type == ext_type and scandata.name == ext_name:
-                is_root = True
-
-            if not is_root:
-                # scan dependencies and save findings to ARI RAM
-                dep_scanner = ARIScanner(
-                    root_dir=self.root_dir,
-                    rules_dir=self.rules_dir,
-                    rules=self.rules,
-                    ram_client=self.ram_client,
-                    read_ram=self.read_ram,
-                    write_ram=self.write_ram,
-                    use_ansible_doc=self.use_ansible_doc,
-                    do_save=self.do_save,
-                    silent=True,
-                )
-                # use prepared dep dirs
-                dep_scanner.evaluate(
-                    type=ext_type,
-                    name=ext_name,
-                    version=ext_ver,
-                    hash=ext_hash,
-                    target_path=ext_path,
-                    dependency_dir=scandata.dependency_dir,
-                    source_repository=scandata.source_repository,
-                )
-
-                if self.read_ram:
-                    # searching findings from ARI RAM and use them if found
-                    loaded, ext_defs = self.load_definitions_from_ram(ext_type, ext_name, ext_ver, ext_hash)
-                    if loaded:
-                        key = "{}-{}".format(ext_type, ext_name)
-                        scandata.ext_definitions[key] = ext_defs
-                        if not self.silent:
-                            logger.debug(f'Use spec data for "{ext_name}" in RAM DB')
-                        continue
-
-                # load the src directory
-                scandata.load_definition_ext(ext_type, ext_name, ext_path)
-        self.record_end(time_records, "dependency_load")
-
-        if not self.silent:
-            logger.debug("load_definition_ext() done")
-
         loaded = False
         self.record_begin(time_records, "target_load")
-        if self.read_ram and scandata.type not in [LoadType.PLAYBOOK, LoadType.TASKFILE]:
+        if self.read_ram and scandata.type not in [LoadType.PLAYBOOK, LoadType.TASKFILE, LoadType.PROJECT]:
             loaded, root_defs = self.load_definitions_from_ram(scandata.type, scandata.name, scandata.version, scandata.hash, allow_unresolved=True)
             logger.debug(f"spec data loaded: {loaded}")
             if loaded:
@@ -914,6 +964,8 @@ class ARIScanner(object):
         if not loaded:
             scandata.load_definitions_root(target_path=scandata.target_path)
 
+        scandata.set_target_object()
+
         if not self.silent:
             logger.debug("load_definitions_root() done")
             playbooks_num = len(scandata.root_definitions["definitions"]["playbooks"])
@@ -929,6 +981,11 @@ class ARIScanner(object):
         if not self.silent:
             logger.debug("apply_spec_rules() done")
 
+        # load_only is True when this scanner is scanning dependency
+        # otherwise, move on tree construction / rule evaluation
+        if load_only:
+            return None
+
         _ram_client = None
         if self.read_ram:
             _ram_client = self.ram_client
@@ -966,15 +1023,20 @@ class ARIScanner(object):
             # print("root definitions:", root_counts)
 
         # save RAM data
-        if self.write_ram:
+        if self.write_ram and scandata.type not in [LoadType.PLAYBOOK, LoadType.TASKFILE, LoadType.PROJECT]:
             self.register_findings_to_ram(scandata.findings)
-            self.register_module_index_to_ram(scandata.findings)
+            self.register_indices_to_ram(scandata.findings, include_test_contents)
 
         if scandata.out_dir is not None and scandata.out_dir != "":
             self.save_rule_result(scandata.findings, scandata.out_dir)
             if not self.silent:
                 print("The rule result is saved at {}".format(scandata.out_dir))
 
+            if objects:
+                self.save_definitions(scandata.root_definitions, scandata.out_dir)
+                if not self.silent:
+                    print("The objects is saved at {}".format(scandata.out_dir))
+
         if not self.silent:
             summary = summarize_findings(scandata.findings, self.show_all)
             print(summary)
@@ -1012,12 +1074,16 @@ class ARIScanner(object):
                     target_path=target_path,
                     dependency_dir=dependency_dir,
                     download_only=download_only,
+                    load_only=load_only,
+                    skip_dependency=skip_dependency,
                     use_src_cache=use_src_cache,
                     source_repository=source_repository,
                     playbook_yaml=playbook_yaml,
                     playbook_only=playbook_only,
                     taskfile_yaml=taskfile_yaml,
                     taskfile_only=taskfile_only,
+                    include_test_contents=include_test_contents,
+                    objects=objects,
                     raw_yaml=raw_yaml,
                     out_dir=out_dir,
                     spec_mutations_from_previous_scan=scandata.spec_mutations,
@@ -1042,8 +1108,8 @@ class ARIScanner(object):
     def register_findings_to_ram(self, findings: Findings):
         self.ram_client.register(findings)
 
-    def register_module_index_to_ram(self, findings: Findings):
-        self.ram_client.module_index_register(findings)
+    def register_indices_to_ram(self, findings: Findings, include_test_contents: bool = False):
+        self.ram_client.register_indices_to_ram(findings, include_test_contents)
 
     def save_findings(self, findings: Findings, out_dir: str):
         self.ram_client.save_findings(findings, out_dir)
@@ -1057,6 +1123,18 @@ class ARIScanner(object):
 
         findings.save_rule_result(fpath=os.path.join(out_dir, "rule_result.json"))
 
+    def save_definitions(self, definitions: dict, out_dir: str):
+        if out_dir == "":
+            raise ValueError("output dir must be a non-empty value")
+
+        if not os.path.exists(out_dir):
+            os.makedirs(out_dir, exist_ok=True)
+
+        objects_json_str = jsonpickle.encode(definitions["definitions"], make_refs=False)
+        fpath = os.path.join(out_dir, "objects.json")
+        with open(fpath, "w") as file:
+            file.write(objects_json_str)
+
     def get_last_scandata(self):
         return self._current
 
