@@ -485,14 +485,20 @@ class TreeLoader(object):
             self.trees.append(tree_objects)
         return self.trees, additional_objects
 
-    def _recursive_get_calls(self, key, caller=None, handover={}, index=0):
+    def _recursive_get_calls(self, key, caller=None, handover={}, index=0, history=[]):
         obj_list = ObjectList()
         obj = self.get_object(key)
         if obj is None:
             return obj_list
+        if key in history:
+            return obj_list
+        _history = []
+        if history:
+            _history = [h for h in history]
         call_obj = call_obj_from_spec(spec=obj, caller=caller, index=index)
         if call_obj is not None:
             obj_list.add(call_obj, update_dict=False)
+            _history.append(key)
         children_keys, from_ram, handover = self._get_children_keys(obj, handover_from_upper_node=handover)
         for i, c_key in enumerate(children_keys):
             child_objects = self._recursive_get_calls(
@@ -500,6 +506,7 @@ class TreeLoader(object):
                 call_obj,
                 handover,
                 i,
+                _history,
             )
             if isinstance(call_obj, TaskCall):
                 taskcall = call_obj
@@ -666,8 +673,10 @@ class TreeLoader(object):
                 tasks_from = handover_from_upper_node.get("tasks_from")
                 if tasks_from:
                     target_taskfiles = [tasks_from]
-            main_taskfile_key = [tf for tf in obj.taskfiles if tf.split(key_delimiter)[-1].split("/")[-1] in target_taskfiles]
-            children_keys.extend(main_taskfile_key)
+            target_taskfile_key = [
+                tf for tf in obj.taskfiles if tf.split(key_delimiter)[-1].split("/")[-1] in target_taskfiles and "/handlers/" not in tf
+            ]
+            children_keys.extend(target_taskfile_key)
         elif isinstance(obj, TaskFile):
             children_keys = obj.tasks
         elif isinstance(obj, Task):
